#include "mbs.h"

/*********************************************************************
**
**   MACROS:        
**
**   ROUTINES:      mbs_cube_root
**                 
**********************************************************************/

/*********************************************************************
**
**   Routine Name:  mbs_cube_root
**                  
**   Purpose:       compute the roots of a cubic polynomial
**                  of the form:
**                      ((a1*x + a2)*x + a3)*x + a4
**
**   Input:         coef: vector of coefficients a1...a4
**                        coef[0]=a1,...,coef[3]=a4
**                  there is no intrinsic check for proper allocation
**                  and initialization of the coefficients --- watch
**                  out!
**                  
**   Output:        the roots
**
**   Called By:     
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

sb_complex *
mbs_cube_root(double *coef)
{
    double       a, b, c, alpha, beta, phi, psi, theta;
    double       aux, pi3;
    sb_complex  *root;
    
    pi3 = 4.0 * atan(1.0) / 3.0;                /*    pi/3    */
    
    if (!(root = (sb_complex *)calloc(3, sizeof(sb_complex))))
	mbs_alloc_error ("MBS CUBE ROOT", "cubic roots");

    a = coef[1] / coef[0] / 3.0;
    b = coef[2] / coef[0] / 3.0;
    c = coef[3] / coef[0];
    
    alpha = a * a - b;
    beta  = a * (2.0 * a * a - 3.0 * b) + c;
    
    if (alpha > 0.0)
    {
	if (beta == 0.0)
	{
	    aux = sqrt(3.0 * alpha);
	    root[0].real = -a;
	    root[1].real = -a - aux;
	    root[2].real = -a + aux;
	    
	    return(root);
	}
	else
	{
	    aux = 2.0 * exp(1.5*log(alpha));      /*  2.0 * alpha^(3/2)  */
	    
	    if (beta > aux)
	    {
		psi = acosh(beta/aux) / 3.0;
		aux = sqrt(alpha) * cosh(psi);
		root[0].real = -2.0 * aux - a;
		root[1].real = root[2].real = aux - a;
		root[1].imag = sqrt(3.0*alpha) * sinh(psi);
		root[2].imag = -(root[1].imag);
		
		return(root);
	    }
	    else if (beta < -aux)
	    {
		psi = acosh(-beta/aux);
		aux = sqrt(alpha) * cosh(psi);
		root[0].real = 2.0 * aux - a;
		root[1].real = root[2].real = -aux - a;
		root[1].imag = sqrt(3.0*alpha) * sinh(psi);
		root[2].imag = -(root[1].imag);
		
		return(root);
	    }
	    else
	    {
		phi = asin(beta/aux) / 3.0;
		aux = 2.0 * sqrt(alpha);
		root[0].real = aux * sin(phi) - a;
		root[1].real = -aux * sin(pi3+phi) - a;
		root[2].real = aux * sin(pi3-phi) - a;
		
		return(root);
	    }
	}
    }
    else if (alpha == 0.0)
    {
	aux = exp(log(beta)/3.0);               /*  beta^(1/3)  */
	root[0].real = -aux - a;
	root[1].real = root[2].real = aux / 2.0 - a;
	root[1].imag = 3.0 * aux * aux / 4.0;
	root[2].imag = -(root[1].imag);
	
	return(root);
    }
    else
    {
	aux = sqrt(-alpha);
	theta = asinh(beta/(2.0*aux*aux*aux)) / 3.0;
	aux *= sinh(theta);
	root[0].real = -2.0 * aux - a;
	root[1].real = root[2].real = aux - a;
	root[1].imag = sqrt(-3.0*alpha) * cosh(theta);
	root[2].imag = -(root[1].imag);
	
	return(root);
    }
}
