/*********************************************************************
**
**   MACROS:        mbs_ini_clams_runge_kutta
**                  mbs_ini_clams_adams_bashforth
**                  mbs_ini_clams_backward_diff
**
**   ROUTINES:      mbs_set_clams_ode_right_hand_side
**                  mbs_ini_clams_ddrive
**                  mbs_set_clams_ddrive_ewt
**                  mbs_set_clams_ddrive_parameter
**                  mbs_ini_clams_depac
**                  mbs_set_clams_depac_info
**                  mbs_set_clams_depac_eps
**                  mbs_set_clams_depac_abs
**                 
**********************************************************************/


#include "mbs.h"

void 
mbs_set_clams_ode_right_hand_side (void (*rhs)(int,double*,double*,double*))
{
    if (rhs)  __ode_RHS__ = rhs;
    else
    {
	fprintf (stderr, "\tMBS SET CLAMS ODE RIGHT HAND SIDE -- WARNING:\n");
	fprintf (stderr, "No right hand side specified.\n");
	fprintf (stderr, "Continuing...\n");
    }
}


void
mbs_ini_clams_ddrive (int dim, void (*rhs)(int,double*,double*,double*))
{
    if (!(__CDVP = (_Clams_DDrive_Parameters *)calloc(1, sizeof(_Clams_DDrive_Parameters))))
	mbs_alloc_error ("MBS INI CLAMS DDRIVE", "Clams_DDrive_Parameters");
    __CDVP->dim = dim;
    __CDVP->n_state = 1;  /* initialize integration (will change and is checked while integrating) */
    __CDVP->n_task = 1;   /* integrate beyond t_out and interpolate                                */
    __CDVP->n_root = 0;   /* no root search                                                        */
    __CDVP->i_error = 3;  /* keep  max(abs(y[i]), ewt[1]) below eps                                */
    __CDVP->m_int = 3;    /* switch between Adams method and Gear method as needed                 */
    __CDVP->m_iter = 2;   /* chord method; do not provide analytic Jacobian                        */
    __CDVP->impl = 0;     /* differential equations only                                           */
    __CDVP->m_u = 0;      
    __CDVP->m_l = 0;          
    __CDVP->max_ord = 12;
    __CDVP->epsilon = 1e-6;
    mbs_set_clams_ode_right_hand_side (rhs);
}


void
mbs_set_clams_ddrive_ewt (int dim, double *array)
{
    int             i;
    static char     lower[32], upper[32], input[32];
    
    if (!__CDVP)
    {
	fprintf (stderr, "MBS SET CLAMS DDRIVE EWT -- ERROR\n");
	fprintf (stderr, "DDRIVE parameters not allocated!\n");
	fprintf (stderr, "Insert call to MBS INI CLAMS DDRIVE.\n");
	fprintf (stderr, "Exiting now.\n\n");
	exit (-5);
    }
    if (__CDVP->ewt)    free(__CDVP->ewt);
    if (!(__CDVP->ewt = (double *)calloc(dim, sizeof(double))))
	mbs_alloc_error ("MBS SET CLAMS DDRIVE EWT", "ewt");
    for (i=0 ; i<dim ; i++)
    {
	if (array[i] < 0.0)
	{
	    sprintf (lower, "0.0"); sprintf (upper, "infinity"); sprintf (input, "%e", array[i]);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE EWT", "EWT", lower, upper, input);
	}
	__CDVP->ewt[i] = array[i];
    }
}


void
mbs_set_clams_ddrive_parameter (int id, void *value)
{
    int            *ivalue;
    double         *dvalue;
    static char     lower[32], upper[32], input[32];
    
    if (!__CDVP)
    {
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER -- ERROR\n");
	fprintf (stderr, "DDRIVE parameters not allocated!\n");
	fprintf (stderr, "Insert call to MBS INI CLAMS DDRIVE.\n");
	fprintf (stderr, "Exiting now.\n\n");
	exit (-5);
    }
    if (id == __CDVP__EPSILON)  dvalue = value;
    else                        ivalue = value;
    switch (id)
    {
      case  __CDVP__N_STATE:
	if (*ivalue != 1)
	{
	    sprintf (lower, "1"); sprintf (upper, "1"); sprintf (input, "%d", *ivalue);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE PARAMETER", "NSTATE", lower, upper, input);
	}
	__CDVP->n_state = *ivalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: NSTATE = %d\n", __CDVP->n_state);
	break;
      case  __CDVP__N_TASK:
	if (*ivalue < 1  ||  *ivalue > 3)
	{
	    sprintf (lower, "1"); sprintf (upper, "3"); sprintf (input, "%d", *ivalue);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE PARAMETER", "NTASK", lower, upper, input);
	}
	__CDVP->n_task = *ivalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: NTASK = %d\n", __CDVP->n_task);
	break;
      case  __CDVP__N_ROOT:
	if (*ivalue < 0  ||  *ivalue > __CDVP->dim)
	{
	    sprintf (lower, "0"); sprintf (upper, "%d", __CDVP->dim); sprintf (input, "%d", *ivalue);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE PARAMETER", "NROOT", lower, upper, input);
	}
	__CDVP->n_root = *ivalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: NROOT = %d\n", __CDVP->n_root);
	break;
      case  __CDVP__I_ERROR:
	if (*ivalue < 1  ||  *ivalue > 5)
	{
	    sprintf (lower, "1"); sprintf (upper, "5"); sprintf (input, "%d", *ivalue);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE PARAMETER", "IERROR", lower, upper, input);
	}
	__CDVP->i_error = *ivalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: IERROR = %d\n", __CDVP->i_error);
	break;
      case  __CDVP__M_INT:
	if (*ivalue < 1  ||  *ivalue > 3)
	{
	    sprintf (lower, "1"); sprintf (upper, "3"); sprintf (input, "%d", *ivalue);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE PARAMETER", "MINT", lower, upper, input);
	}
	__CDVP->m_int = *ivalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: MINT = %d\n", __CDVP->m_int);
	break;
      case  __CDVP__M_ITER:
	if (*ivalue < 0  ||  *ivalue > 5)
	{
	    sprintf (lower, "0"); sprintf (upper, "5"); sprintf (input, "%d", *ivalue);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE PARAMETER", "MITER", lower, upper, input);
	}
	__CDVP->m_iter = *ivalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: MITER = %d\n", __CDVP->m_iter);
	break;
      case  __CDVP__IMPL:
	if (*ivalue < 0  ||  *ivalue > 3)
	{
	    sprintf (lower, "0"); sprintf (upper, "3"); sprintf (input, "%d", *ivalue);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE PARAMETER", "IMPL", lower, upper, input);
	}
	__CDVP->impl = *ivalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: IMPL = %d\n", __CDVP->impl);
	break;
      case  __CDVP__M_U:
	if (*ivalue < 0  ||  *ivalue > __CDVP->dim)
	{
	    sprintf (lower, "0"); sprintf (upper, "%d", __CDVP->dim); sprintf (input, "%d", *ivalue);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE PARAMETER", "MU", lower, upper, input);
	}
	__CDVP->m_u = *ivalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: MU = %d\n", __CDVP->m_u);
	break;
      case  __CDVP__M_L:
	if (*ivalue < 0  ||  *ivalue > __CDVP->dim)
	{
	    sprintf (lower, "0"); sprintf (upper, "%d", __CDVP->dim); sprintf (input, "%d", *ivalue);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE PARAMETER", "ML", lower, upper, input);
	}
	__CDVP->m_l = *ivalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: ML = %d\n", __CDVP->m_l);
	break;
      case  __CDVP__MAX_ORD:
	if (*ivalue < 0  ||  *ivalue > 12)
	{
	    if (__CDVP->m_int == 2)   sprintf (upper, "5");
	    else                      sprintf (upper, "12");
	    sprintf (lower, "0"); sprintf (input, "%d", *ivalue);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE PARAMETER", "MAXORD", lower, upper, input);
	}
	__CDVP->max_ord = *ivalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: MAXORD = %d\n", __CDVP->max_ord);
	break;
      case  __CDVP__EPSILON:
	if (*dvalue < 0.0)
	{
	    sprintf (lower, "0.0"); sprintf (upper, "infinity"); sprintf (input, "%e", *dvalue);
	    mbs_out_of_range_error ("MBS SET CLAMS DDRIVE PARAMETER", "EPS", lower, upper, input);
	}
	__CDVP->epsilon = *dvalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: EPS = %e\n", __CDVP->epsilon);
	break;
      case  __CDVP__DIM:
	__CDVP->dim = *ivalue;
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER: DIM = %d\n", __CDVP->dim);
      default:
	fprintf (stderr, "MBS SET CLAMS DDRIVE PARAMETER -- WARNING\n");
	fprintf (stderr, "Unknown parameter type: %d\n", id);
	fprintf (stderr, "Ignoring request...\n");
	break;
    }
}


void
mbs_ini_clams_depac (int dim, void (*rhs)(int,double*,double*,double*), int type)
{
    if (!(__CDEPACP = (_Clams_DEPAC_Parameters *)calloc(1, sizeof(_Clams_DEPAC_Parameters))))
	mbs_alloc_error ("MBS INI CLAMS DEPAC", "DEPAC PARAMETERS");
    __CDEPACP->dim = dim;
    __CDEPACP->integ_type = type;
    if (!(__CDEPACP->info = (int *)calloc(15, sizeof(int))))
	mbs_alloc_error ("MBS INI CLAMS DEPAC", "INFO");
    switch (type)
    {
      case __CDEPAC__DDERKF:
	__CDEPACP->info[0] = 0;
	__CDEPACP->info[1] = 0;
	__CDEPACP->info[2] = 0;
	break;
      case __CDEPAC__DDEABM:
	__CDEPACP->info[0] = 0;
	__CDEPACP->info[1] = 0;
	__CDEPACP->info[2] = 0;
	__CDEPACP->info[3] = 0;
	break;
      case __CDEPAC__DDEBDF:
	__CDEPACP->info[0] = 0;
	__CDEPACP->info[1] = 0;
	__CDEPACP->info[2] = 0;
	__CDEPACP->info[3] = 0;
	__CDEPACP->info[4] = 0;
	__CDEPACP->info[5] = 0;
	break;
      default:
	fprintf (stderr, "MBS INI CLAMS DEPAC -- ERROR\n");
	fprintf (stderr, "Unknown integration mode: %d\n", type);
	fprintf (stderr, "Exiting now.\n\n");
	exit (-11);
	break;
    }
    mbs_set_clams_ode_right_hand_side (rhs);
}


void
mbs_set_clams_depac_info (int id, int value)
{
    static char     lower[32], upper[32], input[32], info[32];

    if (!__CDEPACP)
    {
	fprintf (stderr, "MBS SET CLAMS DEPAC INFO -- ERROR\n");
	fprintf (stderr, "DEPAC parameters not allocated!\n");
	fprintf (stderr, "Insert call to MBS INI CLAMS DEPAC.\n");
	fprintf (stderr, "Exiting now.\n\n");
	exit (-5);
    }
    if (value < 0  ||  value > 1)
    {
	sprintf (lower, "0"); sprintf (upper, "1"); 
	sprintf (info, "INFO(%d)", id+1); sprintf (input, "%d", value);
	mbs_out_of_range_error ("MBS SET CLAMS DEPAC INFO", info, lower, upper, input);
    }
    if (__CDEPACP->integ_type == __CDEPAC__DDERKF)
    {
	if (id < 0  ||  value > 2)
	{
	    sprintf (lower, "0"); sprintf (upper, "2"); sprintf (input, "%d", id);
	    mbs_out_of_range_error ("MBS SET CLAMS DEPAC INFO", "INDEX", lower, upper, input);
	}
    }
    else if (__CDEPACP->integ_type == __CDEPAC__DDEABM)
    {
	if (id < 0  ||  value > 3)
	{
	    sprintf (lower, "0"); sprintf (upper, "3"); sprintf (input, "%d", id);
	    mbs_out_of_range_error ("MBS SET CLAMS DEPAC INFO", "INDEX", lower, upper, input);
	}
    }
    else if (__CDEPACP->integ_type == __CDEPAC__DDEBDF)
    {
	if (id < 0  ||  value > 5)
	{
	    sprintf (lower, "0"); sprintf (upper, "5"); sprintf (input, "%d", id);
	    mbs_out_of_range_error ("MBS SET CLAMS DEPAC INFO", "INDEX", lower, upper, input);
	}
    }
    __CDEPACP->info[id] = value;
}


void
mbs_set_clams_depac_eps (int id, double value)
{
    int             top_index;
    static char     lower[32], upper[32], input[32], info[32];
    
    if (!__CDEPACP)
    {
	fprintf (stderr, "MBS SET CLAMS DEPAC INFO -- ERROR\n");
	fprintf (stderr, "DEPAC parameters not allocated!\n");
	fprintf (stderr, "Insert call to MBS INI CLAMS DEPAC.\n");
	fprintf (stderr, "Exiting now.\n\n");
	exit (-5);
    }
    if (!__CDEPACP->eps)
    {
	if (__CDEPACP->info[1])
	{
	    if (!(__CDEPACP->eps = (double *)calloc(__CDEPACP->dim, sizeof(double))))
		mbs_alloc_error ("MBS SET CLAMS DEPAC EPS", "relative tolerance");
	}
	else
	{
	    if (!(__CDEPACP->eps = (double *)calloc(1, sizeof(double))))
		mbs_alloc_error ("MBS SET CLAMS DEPAC EPS", "relative tolerance");
	}
    }
    if (value < 0.0)
    {
	sprintf (lower, "0.0"); sprintf (upper, "infinity"); 
	sprintf (info, "EPS(%d)", id+1); sprintf (input, "%e", value);
	mbs_out_of_range_warning ("MBS SET CLAMS DEPAC EPS", info, lower, upper, input);
	return;
    }
    if (__CDEPACP->info[1])  top_index = __CDEPACP->dim - 1;
    else                     top_index = 0;
    if (id > top_index)
    {
	sprintf (lower, "0"); sprintf (upper, "%d", top_index); 
	sprintf (info, "EPS(%d)", id+1); sprintf (input, "%d", id);
	mbs_out_of_range_warning ("MBS SET CLAMS DEPAC EPS", info, lower, upper, input);
    }
    else
	__CDEPACP->eps[id] = value;
}


void
mbs_set_clams_depac_abs (int id, double value)
{
    int             top_index;
    static char     lower[32], upper[32], input[32], info[32];
    
    if (!__CDEPACP)
    {
	fprintf (stderr, "MBS SET CLAMS DEPAC INFO -- ERROR\n");
	fprintf (stderr, "DEPAC parameters not allocated!\n");
	fprintf (stderr, "Insert call to MBS INI CLAMS DEPAC.\n");
	fprintf (stderr, "Exiting now.\n\n");
	exit (-5);
    }
    if (!__CDEPACP->abs)
    {
	if (__CDEPACP->info[1])
	{
	    if (!(__CDEPACP->abs = (double *)calloc(__CDEPACP->dim, sizeof(double))))
		mbs_alloc_error ("MBS SET CLAMS DEPAC ABS", "relative tolerance");
	}
	else
	{
	    if (!(__CDEPACP->abs = (double *)calloc(1, sizeof(double))))
		mbs_alloc_error ("MBS SET CLAMS DEPAC ABS", "relative tolerance");
	}
    }
    if (value < 0.0)
    {
	sprintf (lower, "0.0"); sprintf (upper, "infinity"); 
	sprintf (info, "ABS(%d)", id+1); sprintf (input, "%e", value);
	mbs_out_of_range_warning ("MBS SET CLAMS DEPAC ABS", info, lower, upper, input);
	return;
    }
    if (__CDEPACP->info[1])  top_index = __CDEPACP->dim - 1;
    else                     top_index = 0;
    if (id > top_index)
    {
	sprintf (lower, "0"); sprintf (upper, "%d", top_index); 
	sprintf (info, "ABS(%d)", id+1); sprintf (input, "%d", id);
	mbs_out_of_range_warning ("MBS SET CLAMS DEPAC ABS", info, lower, upper, input);
    }
    else
	__CDEPACP->abs[id] = value;
}


