/*********************************************************************
**
**   MACROS:        
**
**   ROUTINES:      
**                 
**********************************************************************/

#include  "mbs.h"
#include  "nrrunge-kutta.h"

/*********************************************************************
**
**   PURPOSE:       Runge-Kutta code from  Press et al., (1988), 
**                  Numerical Recipes in C, Cambridge University Press. 
**                     
**   MACROS:        
**
**   ROUTINES:      
**                 
**********************************************************************/


/* char *malloc(); RdB Jan 2007 */
void rkqc();
void rk4();

void nrerror(error_text)
char error_text[];
{
  fprintf(stdout,"Numerical Recipes run-time error...\n");
  fprintf(stdout,"%s\n",error_text);
  fprintf(stdout,"...now exiting to system...\n");
  exit(1);
}

double *vector(n) int n;
{
  double *v;

  v=(double *)malloc((unsigned)(n*sizeof(double)));
  if (!v) nrerror("allocation failure in vector()");
  return v;
}

void free_vector(v,n) double *v; int n;
{
  free((char*)v);
}

#define MAXSTP 10000
#define TINY 1.0e-30

void odeint(ystart,nvar,x1,x2,eps,h1,hmin,nok,nbad,derivs)  
                                         /* rdb: no rkqc */
double ystart[],x1,x2,eps,h1,hmin;
int nvar,*nok,*nbad;
void (*derivs)(int,double*,double *,double *); 
{
  int nstp,i;
  double x,hnext,hdid,h;
  double *yscal,*y,*dydx,*vector();
  /* void nrerror(),free_vector(); */
  yscal=vector(nvar);
  y=vector(nvar);
  dydx=vector(nvar);
  x=x1;
  h=(x2 > x1) ? fabs(h1) : -fabs(h1);
  *nok = (*nbad) = 0;
  for (i=0;i<nvar;i++) y[i]=ystart[i];
  for (nstp=1;nstp<=MAXSTP;nstp++) {
    (*derivs)(nvar,&x,y,dydx);
    for (i=0;i<nvar;i++)
      yscal[i]=fabs(y[i])+fabs(dydx[i]*h)+TINY;
    if ((x+h-x2)*(x+h-x1) > 0.0) h=x2-x;
    rkqc(y,dydx,nvar,&x,h,eps,yscal,&hdid,&hnext,derivs);
    if (hdid == h) ++(*nok); else ++(*nbad);
    if ((x-x2)*(x2-x1) >= 0.0) {
      for (i=0;i<nvar;i++) ystart[i]=y[i];
      free_vector(dydx,nvar);
      free_vector(y,nvar);
      free_vector(yscal,nvar);
      return;
    }
    if (fabs(hnext) <= hmin) nrerror("Step size too small in ODEINT");
    h=hnext;
  }
  nrerror("Too many steps in routine ODEINT");
}

#undef MAXSTP
#undef TINY

#define PGROW -0.20
#define PSHRNK -0.25
#define FCOR 0.06666666    /* 1.0/15.0 */
#define SAFETY 0.9
#define ERRCON 6.0e-4

void rkqc(y,dydx,n,x,htry,eps,yscal,hdid,hnext,derivs)
double y[],dydx[],*x,htry,eps,yscal[],*hdid,*hnext;
void (*derivs)(int,double *,double *,double *); 
int n;
{
  int i;
  double xsav,hh,h,temp,errmax;
  double *dysav,*ysav,*ytemp,*vector();
  void nrerror(),free_vector();

  dysav=vector(n);
  ysav=vector(n);
  ytemp=vector(n);
  xsav=(*x);
  for (i=0;i<n;i++) {
    ysav[i]=y[i];
    dysav[i]=dydx[i];
  }
  h=htry;
  for (;;) {
    hh=0.5*h;
    rk4(ysav,dysav,n,xsav,hh,ytemp,derivs);
    *x=xsav+hh;
    (*derivs)(n,x,ytemp,dydx);
    rk4(ytemp,dydx,n,*x,hh,y,derivs);
    *x=xsav+h;
    if (*x == xsav) nrerror("Step size too small in routine RKQC");
    rk4(ysav,dysav,n,xsav,h,ytemp,derivs);
    errmax=0.0;
    for (i=0;i<n;i++) {
      ytemp[i]=y[i]-ytemp[i];
      temp=fabs(ytemp[i]/yscal[i]);
      if (errmax < temp) errmax=temp;
    }
    errmax /= eps;
    if (errmax <= 1.0) {
      *hdid=h;
      *hnext=(errmax > ERRCON ?
        SAFETY*h*exp(PGROW*log(errmax)) : 4.0*h);
      break;
    }
    h=SAFETY*h*exp(PSHRNK*log(errmax));
  }
  for (i=0;i<n;i++) y[i] += ytemp[i]*FCOR;
  free_vector(ytemp,n);
  free_vector(dysav,n);
  free_vector(ysav,n);
}

#undef PGROW
#undef PSHRNK
#undef FCOR
#undef SAFETY
#undef ERRCON

void rk4(y,dydx,n,x,h,yout,derivs)
double y[],dydx[],x,h,yout[];
void (*derivs)(int,double *,double *,double *);
int n;
{
  int i;
  double xh,hh,h6, x_plus_h,*dym,*dyt,*yt,*vector();
  void free_vector();

  dym=vector(n);
  dyt=vector(n);
  yt=vector(n);
  hh=h*0.5;
  h6=h/6.0;
  xh=x+hh;
  for (i=0;i<n;i++) yt[i]=y[i]+hh*dydx[i];
  (*derivs)(n,&xh,yt,dyt);
  for (i=0;i<n;i++) yt[i]=y[i]+hh*dyt[i];
  (*derivs)(n,&xh,yt,dym);
  for (i=0;i<n;i++) {
    yt[i]=y[i]+h*dym[i];
    dym[i] += dyt[i];
  }
  x_plus_h=x+h;
  (*derivs)(n,&x_plus_h,yt,dyt);
  for (i=0;i<n;i++)
    yout[i]=y[i]+h6*(dydx[i]+dyt[i]+2.0*dym[i]);
  free_vector(yt,n);
  free_vector(dyt,n);
  free_vector(dym,n);
}

