/*********************************************************************
**
**   ROUTINES:      mbs_clams_ddrive
**                  mbs_clams_runge_kutta
**                  mbs_clams_adams_bashforth
**                  mbs_clams_backward_diff
**                  mbs_num_rec_runge_kutta
**                  mbs_num_rec_bulirsch_stoer
**                  
**  AUXILIARY ROUTINES:
**                  fdrive_
**                  fde_
**                  fa_
**                  initialize_DEPAC_integrator
**                  
**  DUMMY ROUTINES:
**                  g_
**                  users_
**                  jac_
**                  de_jac_
**                  
**********************************************************************/


#include "mbs.h"

#define  PRECISION_MARGIN   0.999999


/*********************************************************************
**
**   Routine Name:  mbs_clams_ddrive
**                  
**   Purpose:       Initialize and run variable step-size,
**                  stiff/non-stiff ODE solver DDRIV3 (CLAMS)
**
**   Input:         n           number of differential equations
**                  t           independent variable; usually, current time
**                  y           vector of dependent variables
**                  goal_t      integrate until   t = goal_t
**                  initialize (flag)
**                              1:  initialize integrator and integrate
**                              0:  just integrate
**   Output:        t          (= goal_t  or less [if NTASK == 2])
**                  y          (at  t = goal_t  or earlier [if NTASK == 2])
**
**   Called By:     application
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void 
mbs_clams_ddrive(int n, double *t, double *y, double goal_t, int initialize)
{
    int            i, k, m;
    static int     n_de, n_state, n_task, n_root, i_error, i_er_flg, impl, *iwork;
    static int     m_int, m_iter, m_l, m_u, max_ord, max_step, len_work, len_iwork;
    static double  eps, h_max, *ewt, *work;
    double         t_last, t_step, t_out;
    void           fdrive_(int*, double*, double*, double*);
    void           fa_(int*, double*, double*, double*, int*, int*, int*, int*);
    void           users_(double*, double*, double*, double*, double*, double*, double*, 
			  double*, int*, int*, int*, int*);
    void           jac_(int*, double*, double*, double*, int*, int*, int*);
    double         g_(int*, double*, double*, int*);
    extern void    ddriv3_(int*, double*, double*,         /*       don't even look at it         */
			   void(*)(int*, double*, double*, double*), 
			   int*, double*, int*, int*, double*, double*, int*, int*, int*,
			   int*, int*, int*, int*, double*, double*, int*, int*, int*, 
			   void(*)(int*, double*, double*, double*, int*, int*, int*), 
			   void(*)(int*, double*, double*, double*, int*, int*, int*, int*), 
			   int*, int*, double(*)(int*, double*, double*, int*), 
			   void(*)(double*, double*, double*, double*, double*, double*,
				   double*, double*, int*, int*, int*, int*), int*);
    if (!__CDVP)
    {
	fprintf (stderr, "MBS CLAMS DDRIVE -- ERROR\n");
	fprintf (stderr, "DDRIVE parameters not allocated!\n");
	fprintf (stderr, "Insert call to MBS INI CLAMS DDRIVE.\n");
	fprintf (stderr, "Exiting now.\n\n");
	exit (-5);
    }
    if (initialize)
    {
	/*
	 * set integrator's parameters
	 */
	n_de = n;
	n_state = __CDVP->n_state;
	n_task  = __CDVP->n_task;
	n_root  = __CDVP->n_root;
	i_error = __CDVP->i_error;
	m_int   = __CDVP->m_int;
	m_iter  = __CDVP->m_iter;
	impl    = __CDVP->impl;
	if (__CDVP->m_u)   m_u = __CDVP->m_u;
	else               m_u = n;
	if (__CDVP->m_l)   m_l = __CDVP->m_l;
	else               m_l = n;
	max_ord = __CDVP->max_ord;
	len_work = 250 + 2*n_root;
	len_iwork = 50;
	if (m_iter == 0)
	{
	    if (impl > 0)
	    {
		fprintf (stderr, "MBS CLAMS DDRIVE -- ERROR\n");
		fprintf (stderr, "Illegal combination of parameters:\n");
		fprintf (stderr, "MITER = 0 requires IMPL = 0.\n");
		fprintf (stderr, "Your selection: IMPL = %d.\n", impl);
		fprintf (stderr, "Exiting now.\n\n");
		exit (-1);
	    }
	    len_work += ((max_ord+4)*n);
	    if (m_int == 3)   len_iwork += n;
	}
	else if (m_iter == 3)
	{
	    len_work += ((max_ord+4)*n);
	    if (m_int == 3)   len_iwork += n;
	}
	else if (m_iter == 1 || m_iter == 2)
	{
	    len_work += (((max_ord+5)+n)*n);
	    len_iwork += n;
	    if      (impl == 1)    len_work += (n*n);
	    else if (impl == 2)    len_work += n;
	    else if (impl == 3)    len_work += (n*n_de);
	}
	else if (m_iter == 4 || m_iter == 5)
	{
	    m = 2*m_l + m_u + 1;
	    len_work += ((max_ord+5+m)*n);
	    len_iwork += n;
	    if      (impl == 1)    len_work += (m*n);
	    else if (impl == 2)    len_work += n;
	    else if (impl == 3)    len_work += (m*n_de);
	}
	max_step = n * 500;
	
	h_max = goal_t;		/* no restrictions on step size */
	eps = __CDVP->epsilon;
	/*
	 *  allocate and initialize work arrays
	 */
	if (i_error <= 3)
	{
	    if (ewt)   free(ewt);
	    if (!(ewt = (double *)calloc(1,sizeof(double))))
		mbs_alloc_error ("MBS CLAMS DDRIVE", "work array EWT");
	    if (__CDVP->ewt)	*ewt = __CDVP->ewt[0];
	    else                *ewt = 1.0e-09;
	}
	else
	{
	    if (ewt)   free(ewt);
	    if (!(ewt = (double *)calloc(n,sizeof(double))))
		mbs_alloc_error ("MBS CLAMS DDRIVE", "work array EWT");
	    if (__CDVP->ewt)
		for (i=0 ; i<n ; i++)   ewt[i] = __CDVP->ewt[i];
	    else
		for (i=0 ; i<n ; i++)   ewt[i] = 1.0e-09;
	}
	if (work)   free(work);
	if (!(work = (double *)calloc(len_work,sizeof(double))))
	    mbs_alloc_error ("MBS CLAMS DDRIVE", "work array WORK");
	if (iwork)   free(iwork);
	if (!(iwork = (int *)calloc(len_iwork,sizeof(int))))
	    mbs_alloc_error ("MBS CLAMS DDRIVE", "work array IWORK");
	if (!__ode_RHS__)
	{
	    fprintf (stderr, "MBS CLAMS DDRIVE -- ERROR:\n");
	    fprintf (stderr, "No right hand side specified.\n");
	    fprintf (stderr, "Exiting now.\n\n");
	    exit (-9);
	}
    }
    if (n_task == 1)
    {
	t_step = (goal_t - *t)/10.0;
	t_last = PRECISION_MARGIN*goal_t;
	t_out = *t;
	do
	{
	    t_out += t_step;
	    do
	    {
		ddriv3_(&n, t, y, fdrive_, &n_state, &t_out, &n_task, &n_root, &eps, ewt, &i_error,
			&m_int, &m_iter, &impl, &m_l, &m_u, &max_ord, &h_max, work, &len_work, 
			iwork, &len_iwork, jac_, fa_, &n_de, &max_step, g_, users_, &i_er_flg);
	    }
	    while (n_state == 3);
	    
	    if (n_state == 12)
	    {
		fprintf(stderr, "Terminated integration due to ERROR #%d\n", i_er_flg);
		exit(12);
	    }
	    else if (6 <= n_state && n_state <= 10)  exit(2);
	}
	while (*t < t_last);
    }
    else
	ddriv3_(&n, t, y, fdrive_, &n_state, &goal_t, &n_task, &n_root, &eps, ewt, 
		&i_error, &m_int, &m_iter, &impl, &m_l, &m_u, &max_ord, &h_max, work, &len_work, 
		iwork, &len_iwork, jac_, fa_, &n_de, &max_step, g_, users_, &i_er_flg);
    if (n_state == 12)
    {
	fprintf(stderr, "Terminated integration due to ERROR #%d\n", i_er_flg);
	exit(12);
    }
    else if (6 <= n_state && n_state <= 10)  exit(2);
}


/*********************************************************************
**
**   Routine Name:  fdrive_
**                  
**   Purpose:       wrap-around for call to routine evaluating the 
**                  right hand side of the differential equation
**
**   Input:         n       number of differential equations
**                  t       independent variable
**                  y       vector of dependent variables
**   Output:        y_dot   vector of derivatives
**
**   Called By:     mbs_clams_ddrive
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void 
fdrive_(int *n, double *t, double *y, double *y_dot)
{
    (*__ode_RHS__)(*n, t, y, y_dot);
}


/*********************************************************************
**
**   Routine Name:  fa_
**                  
**   Purpose:       Compute array  a  for ODE of the form
**                          a*y' = f(y,t)
**                  Normally  a[i] == 1, for all i.
**                  cf. CLAMDOC
**
**   Input:         n         system size
**                  t         independent variable (unused)
**                  y         vector of dependent variables (unused)
**                  matdim    unused
**                  ml        unused
**                  mu        unused
**                  nde       number of differential equations
**   Output:        a         transformation vector
**
**   Called By:     mbs_clams_ddrive
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void 
fa_(int *n, double *t, double *y, double *a, int *matdim, int *ml, int *mu, int *nde)
{
  int     i;

  for (i=0 ; i<*nde ; i++)
    a[i] = 1.0;
  for (i=*nde ; i<*n ; i++)
    a[i] = 0.0;
}


/*********************************************************************
**
**   Routine Name:  jac_
**                  
**   Purpose:       dummy function for linker
**
**   Input:         
**   Output:        
**
**   Called By:     should never be called
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void 
jac_(int *n, double *t, double *y, double *dfdy, int *matdim, int *ml, int *mu)
{
    mbs_dummy_catch ("MBS CLAMS DDRIVE: JAC_");
}


/*********************************************************************
**
**   Routine Name:  g_
**                  
**   Purpose:       dummy function for linker
**
**   Input:         
**   Output:        
**
**   Called By:     should never be called
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double 
g_(int *n, double *t, double *y, int *iroot)
{
    mbs_dummy_catch ("MBS CLAMS DDRIVE: G_");
}


/*********************************************************************
**
**   Routine Name:  users_
**                  
**   Purpose:       dummy function for linker
**
**   Input:         
**   Output:        
**
**   Called By:     should never be called
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void 
users_(double* y, double *yh, double *ywt, double *save_1, double *save_2, 
       double* t, double *h, double *el, int *impl, int *n, int *nde, int *iflag)
{
    mbs_dummy_catch ("MBS CLAMS DDRIVE: USERS_");
}


/*
 *  module global variables for DEPAC integrators
 */
static int     idid, len_rwork, len_iwork, *info, *iwork, *i_parameters;
static double *eps, *depac_abs, *rwork, *r_parameters;
void           initialize_DEPAC_integrator (int, int, int, int, char*);


/*********************************************************************
**
**   Routine Name:  mbs_clams_runge_kutta
**                  
**   Purpose:       Initialize and run variable step-size
**                  Runge-Kutta-Fehlberg ODE solver DDERKF (CLAMS)
**
**   Input:         n           number of differential equations
**                  t           independent variable; usually, current time
**                  y           vector of dependent variables
**                  goal_t      integrate until   t = goal_t
**                  initialize (flag)
**                              1:  initialize integrator and integrate
**                              0:  just integrate
**   Output:        t          (= goal_t)
**                  y          (at  t = goal_t)
**
**   Called By:     application
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void 
mbs_clams_runge_kutta (int n, double *t, double *y, double goal_t, int initialize)
{
    void           fde_(double*, double*, double*, double*, int*);
    extern void    dderkf_(void(*)(double*, double*, double*, double*, int*), 
			   int*, double*, double*, double*, int*, double*, double*,
			   int*, double*, int*, int*, int*, double*, int*);
    if (!__CDEPACP)
    {
	fprintf (stderr, "MBS CLAMS RUNGE KUTTA -- ERROR\n");
	fprintf (stderr, "DEPAC parameters not allocated!\n");
	fprintf (stderr, "Insert call to MBS INI CLAMS DDRIVE.\n");
	fprintf (stderr, "Exiting now.\n\n");
	exit (-5);
    }
    if (initialize)
	initialize_DEPAC_integrator (n, 33+7*n, 34, 3, "MBS CLAMS RUNGE KUTTA");

    dderkf_(fde_, &n, t, y, &goal_t, info, eps, depac_abs, &idid, rwork, &len_rwork, 
	    iwork, &len_iwork, r_parameters, i_parameters);
    if (idid < 0)
    {
	fprintf(stderr, "\tMBS CLAMS RUNGE KUTTA: Integration failed. Error code %d.\n", idid);
	fprintf(stderr, "\tExiting now.\n");
	exit(-10);
    }
}


/*********************************************************************
**
**   Routine Name:  mbs_clams_adams_bashforth
**                  
**   Purpose:       Initialize and run variable step-size
**                  Adams-Bashforth ODE solver DDEABM (CLAMS)
**
**   Input:         n           number of differential equations
**                  t           independent variable; usually, current time
**                  y           vector of dependent variables
**                  goal_t      integrate until   t = goal_t
**                  initialize (flag)
**                              1:  initialize integrator and integrate
**                              0:  just integrate
**   Output:        t          (= goal_t)
**                  y          (at  t = goal_t)
**
**   Called By:     application
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void 
mbs_clams_adams_bashforth (int n, double *t, double *y, double goal_t, int initialize)
{
    void           fde_(double*, double*, double*, double*, int*);
    extern void    ddeabm_(void(*)(double*, double*, double*, double*, int*), 
			   int*, double*, double*, double*, int*, double*, double*,
			   int*, double*, int*, int*, int*, double*, int*);
    if (!__CDEPACP)
    {
	fprintf (stderr, "MBS CLAMS ADAMS BASHFORTH -- ERROR\n");
	fprintf (stderr, "DEPAC parameters not allocated!\n");
	fprintf (stderr, "Insert call to MBS INI CLAMS DEPAC.\n");
	fprintf (stderr, "Exiting now.\n\n");
	exit (-5);
    }
    if (initialize)
	initialize_DEPAC_integrator (n, 130+21*n, 51, 4, "MBS CLAMS ADAMS BASHFORTH");

    ddeabm_(fde_, &n, t, y, &goal_t, info, eps, depac_abs, &idid, rwork, &len_rwork, 
	    iwork, &len_iwork, r_parameters, i_parameters);
    if (idid < 0)
    {
	fprintf(stderr, "\tMBS CLAMS ADAMS BASHFORTH: Integration failed. Error code %d.\n", idid);
	fprintf(stderr, "\tExiting now.\n");
	exit(-10);
    }
}


/*********************************************************************
**
**   Routine Name:  mbs_clams_backward_diff
**                  
**   Purpose:       Initialize and run variable step-size
**                  Backward Differentiation ODE solver DDERKF (CLAMS).
**                  Suited for stiff equations.
**
**   Input:         n           number of differential equations
**                  t           independent variable; usually, current time
**                  y           vector of dependent variables
**                  goal_t      integrate until   t = goal_t
**                  initialize (flag)
**                              1:  initialize integrator and integrate
**                              0:  just integrate
**   Output:        t          (= goal_t)
**                  y          (at  t = goal_t)
**
**   Called By:     application
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void 
mbs_clams_backward_diff (int n, double *t, double *y, double goal_t, int initialize)
{
    void           fde_(double*, double*, double*, double*, int*);
    void           de_jac_(double*, double*, double*, int*, double*, int*);
    extern void    ddebdf_(void(*)(double*, double*, double*, double*, int*), 
			   int*, double*, double*, double*, int*, double*, double*,
			   int*, double*, int*, int*, int*, double*, int*, 
			   void(*)(double*, double*, double*, int*, double*, int*));
    if (!__CDEPACP)
    {
	fprintf (stderr, "MBS CLAMS BACKWARD DIFFERENTIATION -- ERROR\n");
	fprintf (stderr, "DEPAC parameters not allocated!\n");
	fprintf (stderr, "Insert call to MBS INI CLAMS DEPAC.\n");
	fprintf (stderr, "Exiting now.\n\n");
	exit (-5);
    }
    if (initialize)
	initialize_DEPAC_integrator (n, 250+(10+n)*n, 56*n, 6, 
				     "MBS CLAMS BACKWARD DIFFERENTIATION");
    ddebdf_(fde_, &n, t, y, &goal_t, info, eps, depac_abs, &idid, rwork, &len_rwork, 
	    iwork, &len_iwork, r_parameters, i_parameters, de_jac_);
    if (idid < 0)
    {
	fprintf(stderr, "\tMBS CLAMS BACKWARD DIFFERENTIATION: ");
	fprintf(stderr, "Integration failed. Error code %d.\n", idid);
	fprintf(stderr, "\tExiting now.\n");
	exit(-10);
    }
}


/*********************************************************************
**
**   Routine Name:  initialize_DEPAC_integrator
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void
initialize_DEPAC_integrator (int dim, int lrw, int liw, int info_top, char *identifier)
{
    int     i;
    
    if (dim != __CDEPACP->dim)
    {
	fprintf (stderr, "%s -- WARNING:\n", identifier);
	fprintf (stderr, "Dimension of problem, %d, differs\n", dim);
	fprintf (stderr, "from previously specified value, %d.\n", __CDEPACP->dim);
	fprintf (stderr, "Continuing...\n");
    }
    len_rwork = lrw;
    len_iwork = liw;
    /*
     *  flags
     */
    if (info)   free(info);
    if (!(info = (int *)calloc(15, sizeof(int))))
	mbs_alloc_error (identifier, "work array INFO");
    for (i=0 ; i<info_top ; i++) info[i] = __CDEPACP->info[i];
    /*
     *  tolerances
     */
    if (info[1])
    {
	if (!(eps = (double *)calloc(dim, sizeof(double))))
	    mbs_alloc_error (identifier, "relative tolerance");
	if (!(depac_abs = (double *)calloc(dim, sizeof(double))))
	    mbs_alloc_error (identifier, "absolute tolerance");
	if (__CDEPACP->eps)   
	    for (i=0 ; i<dim ; i++)   eps[i] = __CDEPACP->eps[i];
	else
	    for (i=0 ; i<dim ; i++)   eps[i] = 1e-6;
	if (__CDEPACP->abs)
	    for (i=0 ; i<dim ; i++)   depac_abs[i] = __CDEPACP->abs[i];
	else
	    for (i=0 ; i<dim ; i++)   depac_abs[i] = 1e-9;
    }
    else
    {
	if (!(eps = (double *)calloc(1, sizeof(double))))
	    mbs_alloc_error (identifier, "relative tolerance");
	if (!(depac_abs = (double *)calloc(1, sizeof(double))))
	    mbs_alloc_error (identifier, "absolute tolerance");
	if (__CDEPACP->eps)    eps[0] = __CDEPACP->eps[0];
	else                   eps[0] = 1e-6;
	if (__CDEPACP->abs)    depac_abs[0] = __CDEPACP->abs[0];
	else                   depac_abs[0] = 1e-9;
    }
    /*
     *  work arrays
     */
    if (!(rwork = (double *)calloc(len_rwork, sizeof(double))))
	mbs_alloc_error (identifier, "work array RWORK");
    if (!(iwork = (int *)calloc(len_iwork, sizeof(int))))
	mbs_alloc_error (identifier, "work array IWORK");
    /*
     *  parameters for right hand side.
     *  use 'i_parameters[0] = ipar(1)' to pass system size to fde_ .
     *  don't use any other components of 'i_parameters' and none of
     *  'r_parameters'.
     */
    if (!(i_parameters = (int *)calloc(1, sizeof(int))))
	mbs_alloc_error (identifier, "parameter array");
    i_parameters[0] = dim;
}


/*********************************************************************
**
**   Routine Name:  fde_
**                  
**   Purpose:       wrap-around for call to routine evaluating the 
**                  right hand side of the differential equation
**
**   Input:         n       number of differential equations
**                  t       independent variable
**                  y       vector of dependent variables
**   Output:        y_dot   vector of derivatives
**
**   Called By:     mbs_clams_runge_kutta
**                  mbs_clams_adams_bashforth
**                  mbs_clams_backward_diff
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void  
fde_(double *t, double *y, double *y_dot, double *r, int *i)
{
    (*__ode_RHS__)(*i, t, y, y_dot);
}
    

/*********************************************************************
**
**   Routine Name:  de_jac_
**                  
**   Purpose:       dummy function for linker
**
**   Input:         
**   Output:        
**
**   Called By:     should never be called
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void 
de_jac_(double *t, double *y, double *pd, int *row_size, double *r, int *i)
{
    mbs_dummy_catch ("MBS CLAMS INTEGRATORS: DE_JAC_");
}


void  
mbs_num_rec_runge_kutta(int n, double *t, double *y, double goal_t, int initialize)
{
    int     nok, nbad;
    double  eps, h1, hmin; 
    double  t_out, t_last, t_step;
  
    nok = nbad = 0;
    eps = 1e-3;
    h1 = 0.001;
    hmin = 0.0;
   
    odeint(y, n, *t, goal_t, eps, h1, hmin, &nok, &nbad, __ode_RHS__);
}


void  
mbs_num_rec_bulirsch_stoer(int n, double *t, double *y, double goal_t, int initialize)
{
}
