#include "mbs.h"

/*********************************************************************
**
**   MACROS:        
**
**   ROUTINES:      mbs_romberg_uniform
**                  mbs_rbs_atom
**                  mbs_rbs
**                  mbs_romberg
**                 
**********************************************************************/


/*********************************************************************
**
**   Routine Name:  mbs_romberg_uniform
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double 
mbs_romberg_uniform(double (*f)(double), double a, double b, 
		    double eps, double abs, int *limit, int *error)
{
    double  *tf, h, c, cs, s, tf_old, diff;
    int      i, k, q;
    
    if (!(tf = (double *)calloc(*limit+1,sizeof(double))))
	mbs_alloc_error("MBS ROMBERG UNIFORM", "iterates");
    
    h = (b-a)/2.0;
    tf[0] = h*(f(a)+f(b));
    cs = 2.0*h;
    for (k=1 ; k<=*limit ; k++,h/=2.0,cs/=2.0)
    {
	tf_old = tf[0];
	for (s=0.0,c=a+h ; c<b ; c+=cs)
	    s += f(c);
	tf[k] = tf[k-1]/2.0+s*h;
	for (q=4,i=k-1 ; i>=0 ; i--,q*=4)
	{
	    tf[i] = tf[i+1]+(tf[i+1]-tf[i])/(q-1);
	}
	diff = tf_old-tf[0], tf[0];
	if (fabs(diff) < abs ||
	    fabs(diff/tf[0]) < eps)
	{
	    *error = 0;
	    *limit = k;
	    return(tf[0]);
	}
    }	
    fprintf(stderr, "ROMBERG SUMMATION (UNIFORM):\n");
    fprintf(stderr, "Exceeded maximum number of iterations.\n");
    fprintf(stderr, "Returning last estimate.\n");
    *error = 1;
    return(tf[0]);
}

static double t[4];

/*********************************************************************
**
**   Routine Name:  mbs_rbs_atom
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double 
mbs_rbs_atom(double (*f)(double), double a, double b)
{
    double   h, c, cs, s;
    int      i, k, q, m=3;
    
    h = b-a;
    t[0] = h*(f(a)+f(b))/2.0;
    for (k=1 ; k<=m ; k++)
    {
	h/=2.0;
	cs=2.0*h;
	for (s=0.0,c=a+h ; c<b ; c+=cs)
	    s += f(c);
	t[k] = t[k-1]/2.0+s*h;
	for (q=4,i=k-1 ; i>=0 ; i--,q*=4)
	{
	    t[i] = t[i+1]+(t[i+1]-t[i])/(q-1);
	}
    }	
    return(t[0]);
}


/*********************************************************************
**
**   Routine Name:  mbs_rbs
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double 
mbs_rbs(double (*f)(double), double a, double b, 
	double eps, double abs, int limit, int *error)
{
    double   mbs_rbs_atom(double (*)(double), double, double);
    double   h, s, sm, sn, sp, snext, sdiff, lb, ub;
    int      i;

    h = (a+b)/2.0;
    sn = mbs_rbs_atom(f, a, b);
    sm = mbs_rbs_atom(f, a, h);
    sp = mbs_rbs_atom(f, h, b);
    snext = sm+sp;
    sdiff = sn-snext;
    if (fabs(sdiff) < abs ||
	fabs(sdiff/snext) < eps)
    {
	*error = 0;
	return(snext);
    }
    else if (limit == 0)
    {
	(*error)++;
	return(sn);
    }
    else
    {
	h = (b-a)/4.0;
	limit--;
	for (s=0.0,lb=a,ub=a+h,i=0 ; i<4 ; i++,ub+=h,lb+=h)
	    s += mbs_rbs(f, lb, ub, eps, abs, limit, error);
    }
    return (s);
}


/*********************************************************************
**
**   Routine Name:  mbs_romberg
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double 
mbs_romberg(double (*f)(double), double a, double b, 
	    double eps, double abs, int limit, int *error)
{
    double   mbs_rbs(double (*)(double), double, double, 
		     double, double, int, int*);
    double   sum;
    
    sum = mbs_rbs(f, a, b, eps, abs, limit, error);
    if (*error) 
    {
	fprintf(stderr, "ROMBERG SUMMATION (NON-UNIFORM):\n");
	fprintf(stderr, "Exceeded maximum number of iterations.\n");
	fprintf(stderr, "Returning last estimate.\n");
    }
    return(sum);
}
