/*********************************************************************
**
**   MACROS:        
**                  NONE
**
**   ROUTINES:      
**                  initialize_bound_ligand_computation
**                  compute_bound_ligand_concentration
**                  determine_number_of_crosslinks
**                  compute_signaling_crosslinks
**                  determine_number_of_aggregates
**                  compute_signaling_aggregates
**                  default_receptor_conservation
**                  compute_bound_receptors
**                  increment_concentrations
**                  default_get_arguments
**                  get_ligand_attributes
**                  set_ligand_attributes
**                  get_ligand_variables
**                  adjust_ligand_variables
**                  default_print_usage
**                 
**********************************************************************/

#include  "mbs.h"
#include  "ligand.h"

#define  EPS       1e-10
#define  TOL       1e-10

double   **binomial_coefficients;
int        number_of_binomial_coefficients_allocated;


/*********************************************************************
**
**   Routine Name:  initialize_bound_ligand_computation
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     get_ligand_attributes
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void 
initialize_bound_ligand_computation (int size)
{
    number_of_binomial_coefficients_allocated = size;
    binomial_coefficients = mbs_dalloc(size);
    mbs_binomi(binomial_coefficients, size);
}


/*********************************************************************
**
**   Routine Name:  compute_bound_ligand_concentration
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     determine_number_of_crosslinks
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double *
compute_bound_ligand_concentration (double r, LigandAttributes *L,
				    double *bound_ligand_concentration,
				    double total_ligand_concentration)
{
    double    kr, kf, pk, qk, rk, xf;
    int       i;
    
    if (number_of_binomial_coefficients_allocated < L->N_x)
    {
	fprintf (stderr, "Not enough binomial coefficients available.\n");
	exit (-1);
    }
    if (!bound_ligand_concentration)
	if (!(bound_ligand_concentration = (double *)calloc(L->N_x+1, sizeof(double))))
	    mbs_alloc_error("MBS COMPUTE BOUND LIGAND CONCENTRATION", "concentrations");

    kr = L->kappa_x*r;
    xf = L->N_over_N_x * total_ligand_concentration * L->K;
    if (kr >= 1.0)
    {
	kf = 1.0 + 1.0/kr;
	pk = exp((double)(-L->N_x)*log(kr));
	qk = exp((double)(L->N_x)*log(kf));
	xf /= (L->kappa_x*(L->mfk*pk + L->fk*qk));
    }
    else
    {
	kf = 1.0 + kr;
	pk = 1.0;
	qk = exp((double)L->N_x*log(kf));
	xf /= (L->kappa_x+L->N_over_N_x*L->kappa*(qk-1.0));
    }
    for (rk=pk*kr,i=1 ; i<=L->N_x ; i++,rk*=kr)
	bound_ligand_concentration[i] = xf*binomial_coefficients[L->N_x][i]*rk;
    
    return (bound_ligand_concentration);
}


/*********************************************************************
**
**   Routine Name:  determine_number_of_crosslinks
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     USER
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double
determine_number_of_crosslinks (StandardControlVariables *CV,
				StandardCellAttributes *CA,
				StandardLigandVariables *LV,
				LigandAttributes **LA)
{
    int               eflag, k, i;
    double            r, X=0.0;
    double            compute_signaling_crosslinks (LigandAttributes*, double*,
						    StandardCellAttributes*);
    static double   **X_n;
    
    if (!X_n)
	if (!(X_n = (double **)calloc(CV->number_of_ligands, sizeof(double*))))
	    mbs_alloc_error ("MBS DETERMINE NUMBER OF CROSSLINKS", "bound states");

    r = mbs_1d_bisec_froot(default_receptor_conservation, 
			   0.0, 1.0, EPS, TOL, &eflag);
    if (eflag)        exit(1);

    for (k=0 ; k<CV->number_of_ligands ; k++) 
    {
	X_n[k] = compute_bound_ligand_concentration (r, LA[k], X_n[k], 
						     LV[k].concentration);
	X += compute_signaling_crosslinks (LA[k], X_n[k], CA);
    }
    return (X);
}


/*********************************************************************
**
**   Routine Name:  compute_signaling_crosslinks
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     determine_number_of_crosslinks
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double
compute_signaling_crosslinks (LigandAttributes *L, double *X,
			      StandardCellAttributes *CA)
{
    int        i;
    double     csum=0.0, xsum=0.0;
    
    for (i=1 ; i<=L->N_x ; i++)
	if (i >= CA->signal_size)
	{
	    csum += ((double)(i)*X[i]);
	    xsum += X[i];
	}
    return (csum-xsum);
}


/*********************************************************************
**
**   Routine Name:  determine_number_of_aggregates
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     USER
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double
determine_number_of_aggregates (StandardControlVariables *CV,
				StandardCellAttributes *CA,
				StandardLigandVariables *LV,
				LigandAttributes **LA)
{
    int               eflag, k;
    double            r, X=0.0;
    double            compute_signaling_aggregates (LigandAttributes*, double*,
						    StandardCellAttributes*);
    static double   **X_n;
    
    if (!X_n)
	if (!(X_n = (double **)calloc(CV->number_of_ligands, sizeof(double*))))
	    mbs_alloc_error ("MBS DETERMINE NUMBER OF AGGREGATES", "bound states");

    r = mbs_1d_bisec_froot(default_receptor_conservation, 
			   0.0, 1.0, EPS, TOL, &eflag);
    if (eflag)        exit(1);

    for (k=0 ; k<CV->number_of_ligands ; k++) 
    {
	X_n[k] = compute_bound_ligand_concentration (r, LA[k], X_n[k], 
						     LV[k].concentration);
	X += compute_signaling_aggregates (LA[k], X_n[k], CA);
    }
    return (X);
}


/*********************************************************************
**
**   Routine Name:  compute_signaling_aggregates
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     determine_number_of_aggregates
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double
compute_signaling_aggregates (LigandAttributes *L, double *X,
			      StandardCellAttributes *CA)
{
    int        i;
    double     xsum=0.0;
    
    for (i=1 ; i<=L->N_x ; i++)
	if (i >= CA->signal_size)
	    xsum += X[i];
    return (xsum);
}


/*********************************************************************
**
**   Routine Name:  default_receptor_conservation
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     root search, via determine_number_of_crosslinks
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double 
default_receptor_conservation(double r)
{
    double     s=0.0;
    int        i;
    
    for (i=0 ; i<S.number_of_ligands ; i++)
	s += (compute_bound_receptors(r, ligands[i], L[i].concentration));
    return(r*(1.0+s)-1.0);
}


/*********************************************************************
**
**   Routine Name:  compute_bound_receptors
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     default_receptor_conservation
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double
compute_bound_receptors (double r, LigandAttributes *L, 
			 double total_ligand_concentration)
{
    double   kn, kr, result;
    
    result = (double)(L->N)*L->K*total_ligand_concentration;
    kr = L->kappa_x*r + 1.0;
    if (kr > 1.0) 
    {
	kn = exp((double)(-L->N_x)*log(kr));
	result /= (kr * (L->mfk*kn + L->fk));
    }
    else
    {
	kn = exp((double)(L->N_x)*log(kr));
	result /= (kr * (L->mfk + L->fk*kn) / kn);
    }
    return(result);
}


/*********************************************************************
**
**   Routine Name:  increment_concentrations
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     USER
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void
increment_concentrations (StandardControlVariables *S,
			  StandardLigandVariables **L)
{
    int   k;

    for (k=0 ; k<S->number_of_ligands ; k++)
	if ((*L)[k].vary_concentration)
	    (*L)[k].concentration *= S->step_size;  
}


/*********************************************************************
**
**   Routine Name:  default_get_arguments
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     USER
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void 
default_get_arguments (char **argv, 
		       StandardControlVariables *CV,
		       StandardCellAttributes *CA,
		       StandardLigandVariables **LV,
		       LigandAttributes ***LA )
{
    void       get_ligand_attributes (int, char*, StandardCellAttributes*,
				      LigandAttributes***);
    double     get_ligand_variables (int, char*, StandardLigandVariables**);
    void       adjust_ligand_variables (double, StandardControlVariables*,
					StandardLigandVariables*, LigandAttributes**);
    int        nL;
    double     t;
    
    CV->begin = atof(argv[1]);
    CV->end   = atof(argv[2]);
    CV->steps = atoi(argv[3]);
    CV->step_size = exp(log(CV->end/CV->begin)/(double)(CV->steps - 1));
    CV->number_of_ligands = nL = atoi(argv[8]);
    
    CA->alpha          = atof(argv[4]);
    CA->concentration  = atof(argv[5]);
    CA->receptor_total = atof(argv[6]);
    CA->signal_size    = atoi(argv[7]);
    
    get_ligand_attributes (nL, argv[9], CA, LA);
    t = get_ligand_variables(nL, argv[10], LV);
    adjust_ligand_variables(t, CV, *LV, *LA);
}


/*********************************************************************
**
**   Routine Name:  get_ligand_attributes
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     default_get_arguments
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void
get_ligand_attributes (int n_L, char *filename,
		       StandardCellAttributes *CA,
		       LigandAttributes ***L)
{
    FILE    *LA_FILE;
    int      number_of_ligands, n, n_x, n_t=0, i;
    float    w, k, k_x, k_p, k_m, kxp, kxm;
    
    if (!(LA_FILE = fopen (filename, "r")))
    {
	fprintf (stderr, "Unable to open file: %s\n", filename);
	fprintf (stderr, "Exiting now.\n");
	exit(-2);
    }
    fscanf (LA_FILE, "%d", &number_of_ligands);
    if (number_of_ligands != n_L)
    {
	fprintf (stderr, "File contents of file: %s are incompatible\n", filename);
	fprintf (stderr, "with the requested number of ligands: %d.\n", n_L);
	fprintf (stderr, "Exiting now.\n");
	exit(-3);
    }
    if (!((*L) = (LigandAttributes **)calloc(n_L, sizeof(LigandAttributes*))))
	mbs_alloc_error ("MBS GET LIGAND ATTRIBUTES", "Ligand Attributes");
    for (i=0 ; i<n_L ; i++)
    {
	fscanf (LA_FILE, "%e %d %d %e %e %e %e %e %e\n", 
		&w, &n, &n_x, &k, &k_x, &k_p, &k_m, &kxp, &kxm);
	(*L)[i] = set_ligand_attributes ((*L)[i], n, n_x, k, k_x, k_p, k_m, kxp, kxm,
					 w, CA->receptor_total, CA->alpha, CA->concentration);
	n_t = max(n, n_t);
    }
    fclose (LA_FILE);
    initialize_bound_ligand_computation(n_t);
}


/*********************************************************************
**
**   Routine Name:  set_ligand_attributes
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     get_ligand_attributes
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

LigandAttributes *
set_ligand_attributes (LigandAttributes *l, int N, int N_x, double K, double K_x, 
		       double k_p, double k_m, double kxp, double kxm,
		       double weight, double R_total, double cell_concentration, double alpha)
{
    if (!l)
	if (!(l = (LigandAttributes *)calloc(1,sizeof(LigandAttributes))))
	    mbs_alloc_error ("MBS SET LIGAND ATTRIBUTES", "Ligand Attributes");

    l->weight  = weight;	/* in $kg/l$     */
    l->w_micro = weight*1e3;	/* in $\mu g/ml$ */
    l->w_nano  = weight*1e6;    /* in $ng/ml$    */

    l->N = N;
    l->N_x = N_x;
    l->K = K;
    l->K_x = K_x;
    l->k_plus = k_p;
    l->k_minus = k_m;
    l->k_x_plus = kxp;
    l->k_x_minus = kxm;
    l->kappa = K * alpha * cell_concentration * R_total;
    l->kappa_x = K_x * R_total;
    l->N_over_N_x = (double)N / (double)N_x;
    l->fk = l->N_over_N_x * l->kappa / l->kappa_x;
    l->mfk = 1.0 - l->fk;
    
    return (l);
}


/*********************************************************************
**
**   Routine Name:  get_ligand_variables
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     default_get_arguments
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

double
get_ligand_variables (int n_L, char *filename, StandardLigandVariables **V)
{
    FILE    *LV_FILE;
    int      number_of_ligands, i, v;
    float    c, f;
    double   total;
    
    if (!(LV_FILE = fopen (filename, "r")))
    {
	fprintf (stderr, "Unable to open file: %s\n", filename);
	fprintf (stderr, "Exiting now.\n");
	exit(-2);
    }
    fscanf (LV_FILE, "%d", &number_of_ligands);
    if (number_of_ligands != n_L)
    {
	fprintf (stderr, "File contents of file: %s are incompatible\n", filename);
	fprintf (stderr, "with the requested number of ligands: %d.\n", n_L);
	fprintf (stderr, "Exiting now.\n");
	exit(-3);
    }
    if (!((*V) = (StandardLigandVariables *)calloc(n_L, sizeof(StandardLigandVariables))))
	mbs_alloc_error ("MBS GET LIGAND VARIABLES", "Ligand Variables");
    for (total=0.0,i=0 ; i<n_L ; i++)
    {
	fscanf (LV_FILE, "%e %e %d\n", &c, &f, &v);
	(*V)[i].concentration = c;
	(*V)[i].percentage_in_collection = f;
	(*V)[i].vary_concentration = v;
	if (v)  total += f;
    }
    fclose (LV_FILE);
    return (total);
}

/*********************************************************************
**
**   Routine Name:  adjust_ligand_variables
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     default_get_arguments
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/


void
adjust_ligand_variables (double total, StandardControlVariables *CV,
			 StandardLigandVariables *V, LigandAttributes **A)
{
    int    i;

    for (i=0 ; i<CV->number_of_ligands ; i++)
	if (V[i].vary_concentration)
	    V[i].concentration = CV->begin*V[i].percentage_in_collection/total/A[i]->N;
}

		      
/*********************************************************************
**
**   Routine Name:  default_print_usage
**                  
**   Purpose:       
**
**   Input:         
**   Output:        
**
**   Called By:     USER
**
**   Written By:    Bernhard Sulzer, LANL
**   Last Edited:   
**
**********************************************************************/

void
default_print_usage(char *msg)
{
    fprintf(stderr, "%s", msg);
    fprintf(stderr, "USAGE:\tml -S-begin -S-end -S-nsteps -alpha -B -R_T -g -nL -Lfile -Cfile\n\n");
    fprintf(stderr, "\t\t-S-begin\tlower bound of ligand concentrations\n");
    fprintf(stderr, "\t\t-S-end\t\tupper bound of ligand concentrations\n");
    fprintf(stderr, "\t\t-S-nsteps\tnumber of steps for incrementing S\n");
    fprintf(stderr, "\t\t-alpha\t\tvolume-surface translator (cell type!)\n");
    fprintf(stderr, "\t\t-B\t\tcell density\n");
    fprintf(stderr, "\t\t-RT\t\treceptor density\n");
    fprintf(stderr, "\t\t-g\t\tminimal size of signaling aggregate\n");
    fprintf(stderr, "\t\t-nL\t\tnumber of ligands in collection\n");
    fprintf(stderr, "\t\t-Lfile\t\tname of file containing ligand data\n");
    fprintf(stderr, "\t\t-Cfile\t\tname of file containing initial conditions\n");
    fprintf(stderr, "\n");
    exit(1);
}
